@@ -26,16 +26,16 @@
 #include "lib/generic/trie.h"
 #include "lib/resolve.h"
 #include "lib/rplan.h"
+#include "lib/rules/api.h"
 #include "lib/utils.h"
 
 #include "lib/cache/impl.h"
+#include "lib/cache/top.h"
 
 /* TODO:
  *	- Reconsider when RRSIGs are put in and retrieved from the cache.
  *	  Currently it's always done, which _might_ be spurious, depending
  *	  on how kresd will use the returned result.
- *	  There's also the "problem" that kresd ATM does _not_ ask upstream
- *	  with DO bit in some cases.
  */
 
 
@@ -93,7 +93,7 @@ static int assert_right_version(struct kr_cache *cache)
 					(int)ver, (int)CACHE_VERSION);
 			}
 		}
-		ret = cache_op(cache, clear);
+		ret = cache_op(cache, clear, 0);
 	}
 	/* Rewrite the entry even if it isn't needed.  Because of cache-size-changing
 	 * possibility it's good to always perform some write during opening of cache. */
@@ -109,13 +109,22 @@ static int assert_right_version(struct kr_cache *cache)
 
 int kr_cache_open(struct kr_cache *cache, const struct kr_cdb_api *api, struct kr_cdb_opts *opts, knot_mm_t *mm)
 {
-	if (kr_fails_assert(cache))
+	if (kr_fails_assert(cache && opts))
 		return kr_error(EINVAL);
 	memset(cache, 0, sizeof(*cache));
 	/* Open cache */
 	if (!api)
 		api = kr_cdb_lmdb();
 	cache->api = api;
+
+	const size_t orig_maxsize = opts->maxsize;
+	if (opts->maxsize) {
+		const size_t top_size = kr_cache_top_get_size(opts->maxsize);
+		if (!kr_fails_assert(top_size < opts->maxsize)) {
+			opts->maxsize -= top_size;
+		}
+	}
+
 	int ret = cache->api->open(&cache->db, &cache->stats, opts, mm);
 	if (ret == 0) {
 		ret = assert_right_version(cache);
@@ -126,32 +135,46 @@ int kr_cache_open(struct kr_cache *cache, const struct kr_cdb_api *api, struct k
 		 * LMDB only restricts our env without changing the in-file maxsize.
 		 * That is worked around by reopening (found no other reliable way). */
 		cache->api->close(cache->db, &cache->stats);
-		struct kr_cdb_opts opts2;
-		memcpy(&opts2, opts, sizeof(opts2));
+		struct kr_cdb_opts opts2 = *opts;
 		opts2.maxsize = 0;
 		ret = cache->api->open(&cache->db, &cache->stats, &opts2, mm);
 	}
 
-	char *fpath = kr_absolutize_path(opts->path, "data.mdb");
-	if (kr_fails_assert(fpath)) {
-		/* non-critical, but still */
-		fpath = "<ENOMEM>";
-	} else {
-		kr_cache_emergency_file_to_remove = fpath;
-	}
+	free_const(kr_cache_emergency_file_to_remove);
+	kr_cache_emergency_file_to_remove = kr_absolutize_path(opts->path, "data.mdb");
+	kr_assert(kr_cache_emergency_file_to_remove); // non-critical, but still
 
-	if (ret == 0 && opts->maxsize) {
-		size_t maxsize = cache->api->get_maxsize(cache->db);
-		if (maxsize > opts->maxsize) kr_log_warning(CACHE,
-			"Warning: real cache size is %zu instead of the requested %zu bytes."
-			"  To reduce the size you need to remove the file '%s' by hand.\n",
-			maxsize, opts->maxsize, fpath);
+	size_t maxsize = 0;
+	if (ret == 0) {
+		maxsize = cache->api->get_maxsize(cache->db);
+		if (opts->maxsize && (maxsize > opts->maxsize)) {
+			kr_log_notice(CACHE,
+				"real LMDB cache size is %zu instead of the requested %zu bytes, removing all data.\n",
+				maxsize, opts->maxsize);
+			cache_op(cache, clear, opts->maxsize);
+			maxsize = cache->api->get_maxsize(cache->db);
+		}
 	}
 	if (ret != 0)
 		return ret;
 	cache->ttl_min = KR_CACHE_DEFAULT_TTL_MIN;
 	cache->ttl_max = KR_CACHE_DEFAULT_TTL_MAX;
 	kr_cache_make_checkpoint(cache);
+
+	const char *top_path = kr_absolutize_path(opts->path, "top");
+	if (kr_fails_assert(top_path)) {
+		ret = kr_error(errno);
+	}
+	if (ret == 0) {
+		ret = kr_cache_top_init(&cache->top, top_path, orig_maxsize);
+		free_const(top_path);
+	}
+	if (ret != 0) {
+		cache_op(cache, close);
+		cache->db = NULL;
+		return ret;
+	}
+
 	return 0;
 }
 
@@ -162,12 +185,13 @@ const char *kr_cache_emergency_file_to_remove = NULL;
 
 void kr_cache_close(struct kr_cache *cache)
 {
+	kr_cache_top_deinit(&cache->top);
 	kr_cache_check_health(cache, -1);
 	if (cache_isvalid(cache)) {
 		cache_op(cache, close);
 		cache->db = NULL;
 	}
-	free(/*const-cast*/(char*)kr_cache_emergency_file_to_remove);
+	free_const(kr_cache_emergency_file_to_remove);
 	kr_cache_emergency_file_to_remove = NULL;
 }
 
@@ -177,7 +201,7 @@ int kr_cache_commit(struct kr_cache *cache)
 		return kr_error(EINVAL);
 	}
 	if (cache->api->commit) {
-		return cache_op(cache, commit);
+		return cache_op(cache, commit, true, true);
 	}
 	return kr_ok();
 }
@@ -187,7 +211,7 @@ int kr_cache_clear(struct kr_cache *cache)
 	if (!cache_isvalid(cache)) {
 		return kr_error(EINVAL);
 	}
-	int ret = cache_op(cache, clear);
+	int ret = cache_op(cache, clear, 0);
 	if (ret == 0) {
 		kr_cache_make_checkpoint(cache);
 		ret = assert_right_version(cache);
@@ -239,9 +263,7 @@ int32_t get_new_ttl(const struct entry_h *entry, const struct kr_query *qry,
 		int res_stale = qry->stale_cb(res, owner, type, qry);
 		if (res_stale >= 0) {
 			VERBOSE_MSG(qry, "responding with stale answer\n");
-			/* LATER: Perhaps we could use a more specific Stale
-			 * NXDOMAIN Answer code for applicable responses. */
-			kr_request_set_extended_error(qry->request, KNOT_EDNS_EDE_STALE, "6Q6X");
+			qry->request->stale_accounted = true;
 			return res_stale;
 		}
 	}
@@ -274,8 +296,7 @@ static bool check_dname_for_lf(const knot_dname_t *n, const struct kr_query *qry
 /** Return false on types to be ignored.  Meant both for sname and direct cache requests. */
 static bool check_rrtype(uint16_t type, const struct kr_query *qry/*logging*/)
 {
-	const bool ret = !knot_rrtype_is_metatype(type)
-			&& type != KNOT_RRTYPE_RRSIG;
+	const bool ret = !knot_rrtype_is_metatype(type) || type == KNOT_RRTYPE_ANY;
 	if (!ret && kr_log_is_debug_qry(CACHE, qry)) {
 		auto_free char *type_str = kr_rrtype_text(type);
 		VERBOSE_MSG(qry, "=> skipping RR type %s\n", type_str);
@@ -289,9 +310,6 @@ knot_db_val_t key_exact_type_maypkt(struct key *k, uint16_t type)
 	if (kr_fails_assert(check_rrtype(type, NULL)))
 		return (knot_db_val_t){ NULL, 0 };
 	switch (type) {
-	case KNOT_RRTYPE_RRSIG: /* no RRSIG query caching, at least for now */
-		kr_assert(false);
-		return (knot_db_val_t){ NULL, 0 };
 	/* xNAME lumped into NS. */
 	case KNOT_RRTYPE_CNAME:
 	case KNOT_RRTYPE_DNAME:
@@ -317,19 +335,40 @@ int cache_peek(kr_layer_t *ctx, knot_pkt_t *pkt)
 {
 	struct kr_request *req = ctx->req;
 	struct kr_query *qry = req->current_query;
-	/* We first check various exit-conditions and then call the _real function. */
 
+	if (ctx->state & (KR_STATE_FAIL|KR_STATE_DONE))
+		return ctx->state;
+	/* ATM cache only peeks for qry->sname and that would be useless
+	 * to repeat on every iteration, so disable it from now on.
+	 * Note that xNAME causes a followup kr_query, so cache will get re-tried.
+	 * LATER(optim.): assist with more precise QNAME minimization. */
+	if (qry->flags.CACHE_TRIED && !qry->stale_cb)
+		return ctx->state;
+	qry->flags.CACHE_TRIED = true;
+
+	if (!req->options.PASSTHRU_LEGACY) {
+		int ret = kr_rule_local_data_answer(qry, pkt);
+		if (ret < 0)
+			ctx->state = KR_STATE_FAIL;
+		if (ret != 0)
+			return ctx->state;
+	}
+
+	/* We first check various exit-conditions and then call the _real function. */
 	if (!kr_cache_is_open(&req->ctx->cache)
-	    || ctx->state & (KR_STATE_FAIL|KR_STATE_DONE) || qry->flags.NO_CACHE
-	    || (qry->flags.CACHE_TRIED && !qry->stale_cb)
-	    || !check_rrtype(qry->stype, qry) /* LATER: some other behavior for some of these? */
+	    || !check_rrtype(qry->stype, qry)
 	    || qry->sclass != KNOT_CLASS_IN) {
 		return ctx->state; /* Already resolved/failed or already tried, etc. */
 	}
-	/* ATM cache only peeks for qry->sname and that would be useless
-	 * to repeat on every iteration, so disable it from now on.
-	 * LATER(optim.): assist with more precise QNAME minimization. */
-	qry->flags.CACHE_TRIED = true;
+	/* Only skip cache if (flagged and) resolving the origin query or its xNAME chain,
+	 * not when resolving dependencies such as nameserver records or DNSSEC chain. */
+ 	if (qry->flags.NO_CACHE) {
+		const struct kr_query *q = qry;
+		while (q->cname_parent)
+			q = q->cname_parent;
+		if (!q->parent)
+			return ctx->state;
+	}
 
 	if (qry->stype == KNOT_RRTYPE_NSEC) {
 		VERBOSE_MSG(qry, "=> skipping stype NSEC\n");
@@ -353,7 +392,7 @@ static int stash_rrarray_entry(ranked_rr_array_t *arr, int arr_i,
 /** Stash a single nsec_p.  \return 0 (errors are ignored). */
 static int stash_nsec_p(const knot_dname_t *dname, const char *nsec_p_v,
 			struct kr_cache *cache, uint32_t timestamp, knot_mm_t *pool,
-			const struct kr_query *qry/*logging*/);
+			const struct kr_query *qry/*logging + cache_top*/);
 
 /** The whole .consume phase for the cache module. */
 int cache_stash(kr_layer_t *ctx, knot_pkt_t *pkt)
@@ -379,6 +418,14 @@ int cache_stash(kr_layer_t *ctx, knot_pkt_t *pkt)
 		needs_pkt = true;
 		goto stash_packet;
 	}
+	if (qry->stype == KNOT_RRTYPE_ANY || qry->stype == KNOT_RRTYPE_RRSIG) {
+		/* In case of _ANY, useful records could be expected,
+		 * but let's be careful and only stash the whole packet,
+		 * as behavior of servers for _ANY is underspecified and tricky.
+		 * The effect is basically caching _ANY separately from other types. */
+		needs_pkt = true;
+		goto stash_packet;
+	}
 
 	/* Stash individual records. */
 	ranked_rr_array_t *selected[] = kr_request_selected(req);
@@ -412,7 +459,7 @@ int cache_stash(kr_layer_t *ctx, knot_pkt_t *pkt)
 	for (it = trie_it_begin(nsec_pmap); !trie_it_finished(it); trie_it_next(it)) {
 		stash_nsec_p((const knot_dname_t *)trie_it_key(it, NULL),
 				(const char *)*trie_it_val(it),
-				cache, qry->timestamp.tv_sec, &req->pool, req->current_query);
+				cache, qry->timestamp.tv_sec, &req->pool, qry);
 	}
 	trie_it_free(it);
 	/* LATER(optim.): typically we also have corresponding NS record in the list,
@@ -591,8 +638,9 @@ static ssize_t stash_rrset(struct kr_cache *cache, const struct kr_query *qry,
 	};
 
 	/* Prepare raw memory for the new entry. */
+	size_t whole_val_len = 0;
 	ret = entry_h_splice(&val_new_entry, rank, key, k->type, rr->type,
-				rr->owner, qry, cache, timestamp);
+				rr->owner, qry, cache, timestamp, &whole_val_len);
 	if (ret) return kr_ok(); /* some aren't really errors */
 	if (kr_fails_assert(val_new_entry.data))
 		return kr_error(EFAULT);
@@ -607,6 +655,8 @@ static ssize_t stash_rrset(struct kr_cache *cache, const struct kr_query *qry,
 	rdataset_dematerialize(rds_sigs, eh->data + rr_ssize);
 	if (kr_fails_assert(entry_h_consistent_E(val_new_entry, rr->type)))
 		return kr_error(EINVAL);
+	if (qry) // it's possible to insert outside a request
+		kr_cache_top_access(qry->request, key.data, key.len, whole_val_len, "stash_rrset");
 
 	#if 0 /* Occasionally useful when debugging some kinds of changes. */
 	{
@@ -783,6 +833,8 @@ static int stash_nsec_p(const knot_dname_t *dname, const char *nsec_p_v,
 		VERBOSE_MSG(qry, "=> EL write failed (ret: %d)\n", ret);
 		return kr_ok();
 	}
+	if (qry)
+		kr_cache_top_access(qry->request, key.data, key.len, val.len, "stash_nsec_p");
 	if (log_refresh_by) {
 		VERBOSE_MSG(qry, "=> nsec_p stashed for %s (refresh by %d, hash: %x)\n",
 				log_dname, log_refresh_by, log_hash);
@@ -831,7 +883,8 @@ int kr_cache_insert_rr(struct kr_cache *cache,
 	return (int) written;
 }
 
-static int peek_exact_real(struct kr_cache *cache, const knot_dname_t *name, uint16_t type,
+static int peek_exact_real(struct kr_cache *cache, struct kr_request *req,
+			const knot_dname_t *name, uint16_t type,
 			struct kr_cache_p *peek)
 {
 	if (!check_rrtype(type, NULL) || !check_dname_for_lf(name, NULL)) {
@@ -845,7 +898,11 @@ static int peek_exact_real(struct kr_cache *cache, const knot_dname_t *name, uin
 	knot_db_val_t key = key_exact_type(k, type);
 	knot_db_val_t val = { NULL, 0 };
 	ret = cache_op(cache, read, &key, &val, 1);
-	if (!ret) ret = entry_h_seek(&val, type);
+	size_t whole_val_len = 0;
+	if (!ret) {
+		whole_val_len = val.len; // whole size before seeking
+		ret = entry_h_seek(&val, type);
+	}
 	if (ret) return kr_error(ret);
 
 	const struct entry_h *eh = entry_h_consistent_E(val, type);
@@ -860,12 +917,14 @@ static int peek_exact_real(struct kr_cache *cache, const knot_dname_t *name, uin
 		.raw_data = val.data,
 		.raw_bound = knot_db_val_bound(val),
 	};
+	kr_cache_top_access(req, key.data, key.len, whole_val_len, "peek_exact_real"); // hits only
 	return kr_ok();
 }
-int kr_cache_peek_exact(struct kr_cache *cache, const knot_dname_t *name, uint16_t type,
+int kr_cache_peek_exact(struct kr_cache *cache, struct kr_request *req,
+			const knot_dname_t *name, uint16_t type,
 			struct kr_cache_p *peek)
 {	/* Just wrap with extra verbose logging. */
-	const int ret = peek_exact_real(cache, name, type, peek);
+	const int ret = peek_exact_real(cache, req, name, type, peek);
 	if (false && kr_log_is_debug(CACHE, NULL)) { /* too noisy for usual --verbose */
 		auto_free char *type_str = kr_rrtype_text(type),
 			*name_str = kr_dname_text(name);
